import importlib
import inspect
import json
import sys
import traceback
import warnings

import tests
from grader_lib import strip_class, prune_duplicates, retrieve_imports, Nullwriter, add_decorators, remove_decorators, \
    IllegalFunctionError


# Global nullwriter to suppress print output
nullwriter = Nullwriter()
# Raise warnings as errors
warnings.filterwarnings('error')
#Handin name
HANDIN_NAME = "main"

# Message constants TODO include in exception handler
PARAMETER_TYPE_WRONG = "Parameter '%s' has the wrong type '%s', expected is type '%s'."
PARAMETER_TYPE_MISSING = "Parameter '%s' has no type annotation, expected is type '%s'."
RESULT_TYPE_WRONG = "Return value has the wrong type '%s', expected is type '%s'."
RESULT_TYPE_MISSING = "Return value has no type annotation, expected is type '%s'."

WRONG_EXCEPTION_THROWN = "%s thrown instead of %s for %s"
NO_EXCEPTION_THROWN = "%s not thrown for %s"
# RESULT_NOT_EQUAL = "Result does not match expected %s"


# Create a test that evaluates the given comparsion function
def handle_exception(exception, func_log):
    nullwriter.allow_output(True)
    typ = sys.exc_info()[0]
    context_args = sys.exc_info()[1]
    exception_args = exception.args
    tbs = traceback.extract_tb(sys.exc_info()[2])
    tb = None
    for t in tbs:
        if "/" + HANDIN_NAME + ".py" in t[0]:
            tb = t

    msg = None
    if typ is NameError:
        msg = "%s occured in line %i: name '%s' is not defined." % (strip_class(typ), tb[1], tb[3])
    if typ is SyntaxError:
        if not exception_args: #TODO check for which input the cases occur and maybe adjust error messages
            msg = "%s occured: '%s'." % (strip_class(typ), context_args)
        else:
            l = exception_args[1][3]#.replace("\n", "")
            msg = "%s occured in line %i at char %i: '%s'." % (strip_class(typ), exception_args[1][1], exception_args[1][2], l[exception_args[1][2]-1])
    if typ is IllegalFunctionError:
        msg = "Usage of function '%s()' is not allowed in this exercise." % (".".join(exception_args[0]))
    if typ is ImportError:
        msg = "Import of the modules %s is not allowed. Allowed modules are %s." % (exception_args[0][0],
                                                                                    exception_args[0][1])
    if typ is AttributeError:
        msg = "Function is not implemented in given file. Does it have the proper name?"

    #Handle general cases
    if not msg:
        line = -1
        if tb:
            line = tb[1]
        msg = "%s occured in line %i: %s." % (strip_class(typ), line, context_args) #TODO prevent students from
        # raising custom exceptions as print out
    func_log.append(msg)



def check_module_imports(module_name, allowed_modules):
    f = open(module_name + ".py", "r")
    source = f.read()
    imports = retrieve_imports(source)  # Extract all imports in the file
    not_allowed = []
    for mod in imports["modules"]:
        if mod not in allowed_modules:
            not_allowed.append(mod)
    for mod in imports["functions"]:
        if mod not in allowed_modules:
            not_allowed.append(mod)
    f.close()
    return not_allowed, imports


def check_parameter_types(function, types, func_log):
    raise_exception = False
    argspec = inspect.getfullargspec(function)
    #Check all parameters
    for i, p in enumerate(argspec.args):
        if p not in argspec.annotations:
            func_log.append(PARAMETER_TYPE_MISSING % (p, strip_class(types["in"][i])))
            raise_exception = True
            continue
        annotation = argspec.annotations[p]
        if not issubclass(annotation, types["in"][i]):
            func_log.append(PARAMETER_TYPE_WRONG % (p, strip_class(annotation), strip_class(types["in"][i])))
            raise_exception = True

    #Check the return type
    if not types["out"]:
        return raise_exception
    if "return" not in argspec.annotations:
        func_log.append(RESULT_TYPE_MISSING % types["out"])
        raise_exception = True
        return raise_exception

    if not issubclass(argspec.annotations["return"], types["out"]):
        func_log.append(
            RESULT_TYPE_WRONG % (strip_class(argspec.annotations["return"]), strip_class(types["out"])))
        raise_exception = True

    return raise_exception


def function_loader(module_name, definition, func_log):
    function_name, exercise_nr, types = definition["function_name"], definition["name"], definition["types"]
    module = f_return = imports = None
    try:
        # Check if only allowed modules are imported
        not_allowed, imports = check_module_imports(module_name, tests.allowed_modules)
        if len(not_allowed) > 0:
            arguments = (', '.join(map(str, not_allowed)), ', '.join(map(str, tests.allowed_modules)))
            raise ImportError(arguments)

        # Import module and function
        module = importlib.import_module(module_name)
        function = getattr(module, function_name)
        # Check parameter types
        raise_error = check_parameter_types(function, types, func_log)
        if raise_error:
            function = None  # The messages have already been written to the function log

    except Exception as e:
        handle_exception(e, func_log)
    else:
        # Only return the function if no exception was raised
        f_return = function
    return module, f_return, imports


def compare_result(comparison_function, func_log, rem_decos, result):
    rem_decos()
    equal, message = comparison_function(result)
    # handle different errors in comp func
    if not equal:
        func_log.append(message)
    return equal


def check_return_type(function, func_log, result):
    if not issubclass(type(result), function.__annotations__["return"]):
        func_log.append(RESULT_TYPE_WRONG % (
            strip_class(type(result)), strip_class(function.__annotations__["return"])))
        return False
    return True


# Create a test that expects a given error to be raised
def test_error(func_quad, parameters, expected_error, message, omit_output=True):
    function, add_decos, rem_decos, func_log = func_quad
    try:
        # Add decorators to check if illegal call to functions is done and omit output
        if omit_output:
            nullwriter.allow_output(False)
        add_decos()
        # Execute the function with the given parameters, TODO extend with keyword args
        function(*parameters)
    except expected_error:
        return True
    except Exception as e:
        func_log.append(WRONG_EXCEPTION_THROWN % (strip_class(type(e)), expected_error.__name__, message))
        return False
    finally:
        nullwriter.allow_output(True)
        rem_decos()
    func_log.append(NO_EXCEPTION_THROWN % (expected_error.__name__, message))
    return False


def test_function(func_quad, parameters, comparison_function, omit_output=True):
    function, add_decos, rem_decos, func_log = func_quad
    try:
        # Add decorators to check if illegal call to functions is done and omit output
        if omit_output:
            nullwriter.allow_output(False)
        add_decos()
        # Execute the function with the given parameters, TODO extend with keyword args
        result = function(*parameters)
        if type(result) is type(None):
            return True
        # Check if return type matches given type
        if not check_return_type(function, func_log, result):
            return False
        # Compare the result 
        return compare_result(comparison_function, func_log, rem_decos, result)

    except Exception as e:  # For all exceptions in the function execution
        handle_exception(e, func_log)
        return False
    finally:  # Always reset the logging/checking state
        nullwriter.allow_output(True)
        rem_decos()


if __name__ == '__main__':

    results = {}
    error_log = []

    for t in tests.definitions:  # Iterate over all functions to implement for this assignment
        function_log = []
        module, function, imports = function_loader(HANDIN_NAME, t, function_log)
        results[t["name"]] = 0.0
        if not function:  # If loading of the function did not work, continue with next function
            ex_str = "%s: Function %s() has passed 0 tests which results in a score of 0.0 / %s points." % \
                     (t["name"], t["function_name"], t["score"])
            error_log.append([ex_str, function_log])
            continue

        # Decorate forbidden functions and execute the function
        test_results = t["test_function"]([function, add_decorators(t["forbidden_functions"]), \
                                           remove_decorators(t["forbidden_functions"]), function_log])


        # Determine score of the test
        passed_tests = 0
        for r in test_results:
            if r:
                passed_tests += 1
        score = float(passed_tests) / float(len(test_results)) * float(t["score"])

        nullwriter.allow_output(True)
        ex_str = "%s: Function %s() has passed %i / %i tests which results in a score of %.1f / %s points." % \
                 (t["name"], t["function_name"], passed_tests, len(test_results), score, t["score"])
        error_log.append([ex_str, function_log])
        results[t["name"]] = score

    for t in error_log:
        ph = ["="] * len(t[0])
        print("".join(ph) + "\n" + t[0] + "\n")
        msgs = prune_duplicates(t[1])
        for m in msgs:
            print("\t>>> " + m + "\n")

    scores = {"scores": results}
    print(json.dumps(scores))
